Unit Menu;
{ Alex Chalfin 10/2/97        }
{ achalfin@one.net            }
{ Given to the Public Domain  }

{ A simple pop-up menu system for TMT Pascal.             }
{ Note! even though there are scroll bar options, they    }
{ remain un-implemented. They probably won't ever get done}
{ unless the demand becomes *very* high.                  }

{$R-}
{$S-}
{$Q-}

Interface

Const
  MOUSE_BUTTON_1_PRESSED  = $01;
  MOUSE_BUTTON_2_PRESSED  = $02;
  MOUSE_BUTTON_1_RELEASED = $04;
  MOUSE_BUTTON_2_RELEASED = $08;

Type
  MenuCallback = Procedure(Var Data1 : Longint);

Procedure MenuItem(Name : String; Callback : MenuCallback; ClientData : Longint);
Procedure ToggleMenuItem(Name : String; Callback : MenuCallback; InitialState : Longint);
Procedure SubMenu(Name : String);
Procedure EndSubMenu;
Procedure EndMenu;
Procedure MenuAction(MouseButton, MouseX, MouseY : Longint);
Procedure SetVirtualPage(P : Pointer);
Procedure SetScreenSize(Width, Height : Longint);
Procedure SetMenuColors(Light, Medium, Dark : Byte);
Function  IsInMenu : Boolean;

Implementation

Const
  STATE_IDLE      = 0;
  STATE_INIT      = 1;
  STATE_RUN       = 2;
  STATE_DEINIT    = 3;
  STATE_CALL_INIT = 4;
  STATE_DOWN_ONE  = 5;

  InMenu : Longint = STATE_IDLE;

{$I FONT.INC}

Const
  MAX_SUB_MENUS = 16;       { maximum depth of sub menus           }

  MAX_ITEMS_NO_SCROLL = 12; { maximum items to be displayed before }
                            { a scroll bar is inserted             }

  MENU_TYPE_TOGGLE = $01;
  MENU_TYPE_ITEM   = $02;
  MENU_TYPE_SUB    = $03;
  MENU_TYPE_NULL   = $04;

Type
  pMenuSelection = ^tMenuSelection;
  tMenuSelection = Record
    Name        : String;
    Kind        : Longint;
    Callback    : MenuCallback;
    Client      : Longint;
    SubMenuHead : pMenuSelection;
    SubMenuTail : pMenuSelection;
    Next        : pMenuSelection;
  End;

  MenuStackEntry = Record
    Head : pMenuSelection;
    Tail : pMenuSelection;
    x1, y1 : Longint;
    x2, y2 : Longint;
    Flag   : Boolean;
  End;

Var
  ScreenWidth  : Longint;
  ScreenHeight : Longint;
  MenuStack    : Array[0..MAX_SUB_MENUS - 1] of MenuStackEntry;
  MenuStackTop : Longint;

  MenuLightColor  : Byte;
  MenuMediumColor : Byte;
  MenuDarkColor   : Byte;

  MenuStartX      : Longint;
  MenuStartY      : Longint;

  VirtualPage     : Pointer;
  MouseCoordX     : Longint;
  MouseCoordY     : Longint;

Function FindMaxWidth(Index : Longint) : Longint;    Forward;
Function CountMenuItems(Index : Longint): Longint;  Forward;

Procedure SolidRectangle(x1, y1, x2, y2 : Longint; Color : Byte); Forward;
Procedure VerticalLine(x, y1, y2 : Longint; Color : Byte);        Forward;
Procedure HorizontalLine(x1, x2, y : Longint; Color : Byte);      Forward;

Procedure WriteString(X, Y : Longint; S : String; Color : Byte);  Forward;

Procedure DrawMenu; Forward;
Function MouseInBox(x1, y1, x2, y2 : Longint) : Boolean; Forward;
Procedure RecordAction(Button : Longint); Forward;
Function ExecMenuCommand : Longint;       Forward;

Function  IsInMenu : Boolean;

Begin
  IsInMenu := (InMenu = STATE_RUN);
End;

Procedure SetMenuColors(Light, Medium, Dark : Byte);

Begin
  MenuLightColor  := Light;
  MenuMediumColor := Medium;
  MenuDarkColor   := Dark;
End;

Procedure SetVirtualPage(P : Pointer);

Begin
  VirtualPage := P;
End;

Procedure SetScreenSize(Width, Height : Longint);

Begin
  ScreenWidth  := Width;
  ScreenHeight := Height;
End;

Procedure MenuItem(Name : String; Callback : MenuCallback; ClientData : Longint);

Var
  TempPtr  : pMenuSelection;
  Traverse : pMenuSelection;

Begin
  New(TempPtr);

  TempPtr^.Name    := Name;
  TempPtr^.Callback:= Callback;
  TempPtr^.Client  := ClientData;
  TempPtr^.Next    := Nil;
  TempPtr^.SubMenuHead := Nil;
  TempPtr^.SubMenuTail := Nil;

  If (MenuStack[MenuStackTop].Head = Nil)
    Then Begin
      MenuStack[MenuStackTop].Head := TempPtr;
      MenuStack[MenuStackTop].Tail := TempPtr;
    End
    Else Begin
      MenuStack[MenuStackTop].Tail^.Next := TempPtr;
      MenuStack[MenuStackTop].Tail       := MenuStack[MenuStackTop].Tail^.Next;
    End;

  MenuStack[MenuStackTop].Tail^.Kind := MENU_TYPE_ITEM;
End;

Procedure ToggleMenuItem(Name : String; Callback : MenuCallback; InitialState : Longint);

Begin
  MenuItem(Name, Callback, InitialState);
  MenuStack[MenuStackTop].Tail^.Kind := MENU_TYPE_TOGGLE;
End;

Procedure SubMenu(Name : String);

Begin
  MenuItem(Name, Nil, 0);
  MenuStack[MenuStackTop].Tail^.Kind := MENU_TYPE_SUB;

  MenuStackTop := MenuStackTop + 1;
  MenuStack[MenuStackTop].Head := Nil;
  MenuStack[MenuStackTop].Tail := Nil;
End;

Procedure EndSubMenu;

Begin
  MenuStack[MenuStackTop - 1].Tail^.SubMenuHead := MenuStack[MenuStackTop].Head;
  MenuStack[MenuStackTop - 1].Tail^.SubMenuTail := MenuStack[MenuStackTop].Tail;

  MenuStackTop := MenuStackTop - 1;
End;

Procedure EndMenu;

Begin
  MenuStackTop := 0;
End;

Procedure MenuAction(MouseButton, MouseX, MouseY : Longint);

Begin
  MouseCoordX := MouseX;
  MouseCoordY := MouseY;

  RecordAction(MouseButton);

  if (InMenu = STATE_RUN)
    Then DrawMenu;
End;

Procedure RecordAction(Button : Longint);

Var
  i : Longint;

Begin
  if (InMenu = STATE_IDLE) and ((Button and MOUSE_BUTTON_2_PRESSED) > 0)
    Then Begin
      InMenu := STATE_INIT;
      MenuStack[0].Flag := False;
      Exit;
    End;

  if (InMenu = STATE_INIT) and ((Button and MOUSE_BUTTON_2_RELEASED) > 0)
    Then Begin
      MenuStartX := MouseCoordX;
      MenuStartY := MouseCoordY;
      InMenu := STATE_RUN;
      Exit;
    End;

  if (InMenu = STATE_RUN) and ((Button and MOUSE_BUTTON_2_PRESSED) > 0)
    Then Begin
      if (MenuStackTop = 0)
        Then InMenu := STATE_DEINIT
        Else Begin
          InMenu :=  STATE_DOWN_ONE;
          MenuStack[MenuStackTop].Flag := False;
          MenuStackTop := MenuStackTop - 1;
        End;
      Exit;
    End;

  if (InMenu = STATE_DOWN_ONE) and ((Button and MOUSE_BUTTON_2_RELEASED) > 0)
    Then Begin
      InMenu := STATE_RUN;
      Exit;
    End;

  if (InMenu = STATE_DEINIT) and ((Button and MOUSE_BUTTON_2_RELEASED) > 0)
    Then Begin
      InMenu := STATE_IDLE;
      Exit;
    End;

  if (InMenu = STATE_RUN) and ((Button and MOUSE_BUTTON_1_PRESSED) > 0)
    Then Begin
      InMenu := STATE_CALL_INIT;
      Exit;
    End;

  if (InMenu = STATE_CALL_INIT) and ((Button and MOUSE_BUTTON_1_RELEASED) > 0)
    Then Begin
      MenuStartX := MouseCoordX;
      MenuStartY := MouseCoordY;

      i := ExecMenuCommand;
      if (i = MENU_TYPE_ITEM) or (i = MENU_TYPE_TOGGLE)
        Then Begin
          MenuStackTop := 0;
          InMenu := STATE_IDLE;
        End
        Else InMenu := STATE_RUN;

      Exit;
    End;
End;

Function ExecMenuCommand : Longint;

Var
  Traverse : pMenuSelection;
  Index    : Longint;
  i        : Longint;
  RetCode  : Longint;

Begin
  if MouseInBox(MenuStack[MenuStackTop].x1, MenuStack[MenuStackTop].y1,
                MenuStack[MenuStackTop].x2, MenuStack[MenuStackTop].y2)
    Then Begin
      Index := (MouseCoordY - MenuStack[MenuStackTop].y1) Div 11;

      Traverse := MenuStack[MenuStackTop].Head;

      For i := 0 to (Index - 1) do
        Traverse := Traverse^.Next;

      case Traverse^.Kind of
        MENU_TYPE_TOGGLE : Begin
          RetCode := MENU_TYPE_TOGGLE;
          Traverse^.Callback(Traverse^.Client);
        End;
        MENU_TYPE_ITEM : Begin
          RetCode := MENU_TYPE_ITEM;
          Traverse^.Callback(Traverse^.Client);
        End;
        MENU_TYPE_SUB : Begin
          RetCode := MENU_TYPE_SUB;
          MenuStackTop := MenuStackTop + 1;
          MenuStack[MenuStackTop].Head := Traverse^.SubMenuHead;
          MenuStack[MenuStackTop].Tail := Traverse^.SubMenuTail;
          MenuStack[MenuStackTop].Flag := False;
        End;
      End;

      ExecMenuCommand := RetCode;
    End
    Else ExecMenuCommand := MENU_TYPE_NULL;
End;


Procedure DrawMenu;

Var
  i, j           : Longint;
  NumItemsInMenu : Longint;
  WidthMenu      : Longint;
  HeightMenu     : Longint;
  HeightLoop     : Longint;
  Traverse       : pMenuSelection;
  MenuDrawX      : Longint;
  MenuDrawY      : Longint;

Begin
  For i := 0 to MenuStackTop do
    Begin
      NumItemsInMenu := CountMenuItems(i);
      WidthMenu    := FindMaxWidth(i) + 5;

      HeightMenu     := NumItemsInMenu;

      if (NumItemsInMenu > MAX_ITEMS_NO_SCROLL)
        Then Begin
          HeightMenu := MAX_ITEMS_NO_SCROLL;
          WidthMenu  := WidthMenu + 5; { add scrollbar width }
        End;

      HeightLoop := HeightMenu;
      HeightMenu := HeightMenu * 11 + 1; { each item in the menu is 10 pixels high }



      if Not(MenuStack[i].Flag)
        Then Begin
          if ((MenuStartX + WidthMenu) >= ScreenWidth)
            Then MenuDrawX := (ScreenWidth - WidthMenu - 1)
            Else MenuDrawX := MenuStartX;

          if ((MenuStartY + HeightMenu) >= ScreenHeight)
            Then MenuDrawY := (ScreenHeight - HeightMenu - 1)
            Else MenuDrawY := MenuStartY;

          MenuStack[i].x1 := MenuDrawX;
          MenuStack[i].y1 := MenuDrawY;
          MenuStack[i].x2 := MenuDrawX + WidthMenu;
          MenuStack[i].y2 := MenuDrawY + HeightMenu - 1;

          MenuStack[i].Flag := True;
        End
        Else Begin
          MenuDrawX := MenuStack[i].x1;
          MenuDrawY := MenuStack[i].y1;
        End;



      { lets finally draw the menu }
      SolidRectangle(MenuDrawX, MenuDrawY, MenuDrawX + WidthMenu, MenuDrawY + HeightMenu - 1, MenuMediumColor);

      Traverse := MenuStack[i].Head;
      For j := 0 to (HeightLoop - 1) do
        Begin
          if (i = MenuStackTop) and (MouseInBox(MenuDrawX, MenuDrawY + j*11, MenuDrawX + WidthMenu, MenuDrawY + j*11 + 10))
            Then Begin
              case Traverse^.Kind of
                MENU_TYPE_TOGGLE : Begin
                  if (Traverse^.Client = 0)
                    Then WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #133, MenuDarkColor)
                    Else WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #132, MenuDarkColor);
                End;
                MENU_TYPE_ITEM   : Begin
                  WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name, MenuDarkColor);
                End;
                MENU_TYPE_SUB    : Begin
                  WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #131, MenuDarkColor);
                End;
              End;
            End
            Else Begin
              VerticalLine(MenuDrawX, MenuDrawY + j*11, MenuDrawY + j*11 + 10, MenuLightColor);
              VerticalLine(MenuDrawX + WidthMenu - 1, MenuDrawY + j*11, MenuDrawY + j*11 + 10, MenuDarkColor);

              HorizontalLine(MenuDrawX, MenuDrawX + WidthMenu, MenuDrawY + j*11, MenuLightColor);
              HorizontalLine(MenuDrawX, MenuDrawX + WidthMenu, MenuDrawY + j*11 + 10, MenuDarkColor);

              case Traverse^.Kind of
                MENU_TYPE_TOGGLE : Begin
                  if (Traverse^.Client = 0)
                    Then Begin
                      WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #133, MenuLightColor);
                      WriteString(MenuDrawX + 1 + 2 + 1, MenuDrawY + 2 + j*11 + 1, Traverse^.Name + ' ' + #133, MenuDarkColor);
                    End
                    Else Begin
                      WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #132, MenuLightColor);
                      WriteString(MenuDrawX + 1 + 2 + 1, MenuDrawY + 2 + j*11 + 1, Traverse^.Name + ' ' + #132, MenuDarkColor);
                    End;
                End;
                MENU_TYPE_ITEM   : Begin
                  WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name, MenuLightColor);
                  WriteString(MenuDrawX + 1 + 2 + 1, MenuDrawY + 2 + j*11 + 1, Traverse^.Name, MenuDarkColor);
                End;
                MENU_TYPE_SUB    : Begin
                  WriteString(MenuDrawX + 1 + 2, MenuDrawY + 2 + j*11, Traverse^.Name + ' ' + #131, MenuLightColor);
                  WriteString(MenuDrawX + 1 + 2 + 1, MenuDrawY + 2 + j*11 + 1, Traverse^.Name + ' ' + #131, MenuDarkColor);
                End;
              End;
          End;

          Traverse := Traverse^.Next;
        End;

      End;

End;

Function FindMaxWidth(Index : Longint) : Longint;

Var
  Traverse : pMenuSelection;
  MaxWidth : Longint;
  Width    : Longint;

Begin
  MaxWidth := 0;
  Traverse := MenuStack[Index].Head;

  While (Traverse <> Nil) do
    Begin
      case Traverse^.Kind of
        MENU_TYPE_TOGGLE : Width := Length(Traverse^.Name + ' o');
        MENU_TYPE_ITEM   : Width := Length(Traverse^.Name);
        MENU_TYPE_SUB    : Width := Length(Traverse^.Name + ' >');
      End;

      Width := Width * 5;  { width in pixels }

      if (Width > MaxWidth)
        Then MaxWidth := Width;

      Traverse := Traverse^.Next;
    End;

  FindMaxWidth := MaxWidth;
End;

Function CountMenuItems(Index : Longint) : Longint;

Var
  Traverse : pMenuSelection;
  Count    : Longint;

Begin
  Count    := 0;
  Traverse := MenuStack[Index].Head;

  While (Traverse <> Nil) do
    Begin
      Count    := Count + 1;
      Traverse := Traverse^.Next;
    End;

  CountMenuItems := Count;
End;

Procedure SolidRectangle(x1, y1, x2, y2 : Longint; Color : Byte);

Var
  vOfs : DWord;

Begin
  If (x1 >= x2)
    Then Exit;

  vOfs := Ofs(VirtualPage^);
  Inc(vOfs, y1 * ScreenWidth + x1);

  While (y1 < y2) do
    Begin
      FillChar(Mem[vOfs], x2-x1, Color);
      Inc(vOfs, ScreenWidth);
      Inc(y1);
    End;
End;

Procedure VerticalLine(x, y1, y2 : Longint; Color : Byte);

Var
  vOfs : DWord;

Begin
  vOfs := Ofs(VirtualPage^);
  Inc(vOfs, y1 * ScreenWidth + x);

  While (y1 < y2) do
    Begin
      Mem[vOfs] := Color;
      Inc(vOfs, ScreenWidth);
      Inc(y1);
    End;
End;

Procedure HorizontalLine(x1, x2, y : Longint; Color : Byte);

Var
  vOfs : DWord;

Begin
  If (x1 >= x2)
    Then Exit;

  vOfs := Ofs(VirtualPage^);
  Inc(vOfs, y * ScreenWidth + x1);

  FillChar(Mem[vOfs], x2-x1, Color);
End;

Procedure WriteString(X, Y : Longint; S : String; Color : Byte);

Var
  CharCount : Longint;
  i : Longint;
  CharIndex  : Longint;
  xc, yc : Longint;
  sx, sy : Longint;
  Offset : DWord;

Begin
  Offset := Ofs(VirtualPage^);
  For CharCount := 1 to Length(S) do
    Begin
      CharIndex := Ord(S[CharCount]) - 33;
      If (CharIndex > 0) And (CharIndex < NUMCHAR)
        Then Begin
          For yc := 0 to 6 do
            For xc := 0 to 3 do
              If CharSet[CharIndex * 4*7 + yc Shl 2 + xc] = 1
                Then Begin
                  sx := x+xc;
                  sy := y+yc;
                  If ((sy >= 0) and (sy <= ScreenHeight)) And
                     ((sx >= 0) and (sx <= ScreenWidth))
                     Then Mem[DWord(Offset + sy * ScreenWidth + sx)] := Color;
                End;
        End;
      Inc(x, 5);
      If (X > ScreenWidth)
        Then Exit;
    End;
End;

Function MouseInBox(x1, y1, x2, y2 : Longint) : Boolean;

Begin
  if (x1 <= MouseCoordX) and (MouseCoordX <= x2)
    Then Begin
      if (y1 <= MouseCoordY) and (MouseCoordY <= y2)
        Then MouseInBox := True
        Else MouseInBox := False;
    End
    Else MouseInBox := False;
End;



Begin
  MenuStack[0].Head := Nil;
  MenuStack[0].Tail := Nil;
  MenuStackTop      := 0;
End.